@@ -3,6 +3,7 @@
 
 """
 
+import re
 import copy
 import binascii
 import string
@@ -19,7 +20,7 @@ from elftools.elf.constants import SH_FLAGS
 
 
 __author__ = 'Erik Moqvist'
-__version__ = '17.10.3'
+__version__ = '17.14.5'
 
 
 DEFAULT_WORD_SIZE_BITS = 8
@@ -293,6 +294,22 @@ def pretty_ti_txt(line):
     return line
 
 
+def comment_remover(text):
+    def replacer(match):
+        s = match.group(0)
+
+        if s.startswith('/'):
+            return " " # note: a space and not an empty string
+        else:
+            return s
+
+    pattern = re.compile(
+        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
+        re.DOTALL | re.MULTILINE)
+
+    return re.sub(pattern, replacer, text)
+
+
 def is_srec(records):
     try:
         unpack_srec(records.partition('\n')[0].rstrip())
@@ -313,9 +330,20 @@ def is_ihex(records):
 
 def is_ti_txt(data):
     try:
-        return data[0] in ['@', 'q']
-    except IndexError:
+        BinFile().add_ti_txt(data)
+    except Exception:
         return False
+    else:
+        return True
+
+
+def is_verilog_vmem(data):
+    try:
+        BinFile().add_verilog_vmem(data)
+    except Exception:
+        return False
+    else:
+        return True
 
 
 class _Segment:
@@ -841,6 +869,8 @@ class BinFile:
             self.add_ihex(data, overwrite)
         elif is_ti_txt(data):
             self.add_ti_txt(data, overwrite)
+        elif is_verilog_vmem(data):
+            self.add_verilog_vmem(data, overwrite)
         else:
             raise UnsupportedFileFormatError()
 
@@ -971,6 +1001,46 @@ class BinFile:
         if not eof_found:
             raise Error("missing file terminator")
 
+    def add_verilog_vmem(self, data, overwrite=False):
+        address = None
+        chunk = b''
+        words = re.split(r'\s+', comment_remover(data).strip())
+        word_size_bytes = None
+
+        for word in words:
+            if not word.startswith('@'):
+                length = len(word)
+
+                if (length % 2) != 0:
+                    raise Error('Invalid word length.')
+
+                length //= 2
+
+                if word_size_bytes is None:
+                    word_size_bytes = length
+                elif length != word_size_bytes:
+                    raise Error(
+                        f'Mixed word lengths {length} and {word_size_bytes}.')
+
+        for word in words:
+            if word.startswith('@'):
+                if address is not None:
+                    self._segments.add(_Segment(address,
+                                                address + len(chunk),
+                                                chunk,
+                                                self.word_size_bytes))
+
+                address = int(word[1:], 16) * word_size_bytes
+                chunk = b''
+            else:
+                chunk += bytes.fromhex(word)
+
+        if address is not None and chunk:
+            self._segments.add(_Segment(address,
+                                        address + len(chunk),
+                                        chunk,
+                                        self.word_size_bytes))
+
     def add_binary(self, data, address=0, overwrite=False):
         """Add given data at given address. Set `overwrite` to ``True`` to
         allow already added data to be overwritten.
@@ -995,20 +1065,32 @@ class BinFile:
 
         for segment in elffile.iter_segments():
             if segment['p_type'] != 'PT_LOAD':
-                continue
+                 continue
 
-            address = segment['p_paddr']
-            offset = segment['p_offset']
-            size = segment['p_filesz']
+            segment_address = segment['p_paddr']
+            segment_offset = segment['p_offset']
+            segment_size = segment['p_filesz']
 
-            if size == 0:
-                continue
+            for section in elffile.iter_sections():
+                offset = section['sh_offset']
+                size = section['sh_size']
+                address = segment_address + offset - segment_offset
 
-            self._segments.add(_Segment(address,
-                                        address + size,
-                                        data[offset:offset + size],
-                                        self.word_size_bytes),
-                               overwrite)
+                if size == 0:
+                    continue
+
+                if segment_offset <= offset < segment_offset + segment_size:
+                    if section['sh_type'] == 'SHT_NOBITS':
+                        continue
+
+                    if (section['sh_flags'] & SH_FLAGS.SHF_ALLOC) == 0:
+                        continue
+
+                    self._segments.add(_Segment(address,
+                                                address + size,
+                                                data[offset:offset + size],
+                                                self.word_size_bytes),
+                                       overwrite)
 
     def add_file(self, filename, overwrite=False):
         """Open given file and add its data by guessing its format. The format
@@ -1053,6 +1135,15 @@ class BinFile:
         with open(filename, 'r') as fin:
             self.add_ti_txt(fin.read(), overwrite)
 
+    def add_verilog_vmem_file(self, filename, overwrite=False):
+        """Open given Verilog VMEM file and add its contents. Set `overwrite` to
+        ``True`` to allow already added data to be overwritten.
+
+        """
+
+        with open(filename, 'r') as fin:
+            self.add_verilog_vmem(fin.read(), overwrite)
+
     def add_binary_file(self, filename, address=0, overwrite=False):
         """Open given binary file and add its contents. Set `overwrite` to
         ``True`` to allow already added data to be overwritten.
@@ -1266,6 +1357,35 @@ class BinFile:
 
         return '\n'.join(lines) + '\n'
 
+    def as_verilog_vmem(self):
+        """Format the binary file as a Verilog VMEM file and return it as a string.
+
+        >>> print(binfile.as_verilog_vmem())
+
+        """
+
+        lines = []
+
+        if self._header is not None:
+            lines.append(f'/* {self.header} */')
+
+        for segment in self._segments:
+            for address, data in segment.chunks(32 // self.word_size_bytes):
+                words = []
+
+                for i in range(0, len(data), self.word_size_bytes):
+                    word = ''
+
+                    for byte in data[i:i + self.word_size_bytes]:
+                        word += f'{byte:02X}'
+
+                    words.append(word)
+
+                data_hex = ' '.join(words)
+                lines.append(f'@{address:08X} {data_hex}')
+
+        return '\n'.join(lines) + '\n'
+
     def as_binary(self,
                   minimum_address=None,
                   maximum_address=None,
@@ -1654,7 +1774,7 @@ def _convert_input_format_type(value):
                     f"invalid binary address '{items[1]}'")
 
         args = (address, )
-    elif fmt in ['ihex', 'srec', 'auto', 'ti_txt', 'elf']:
+    elif fmt in ['ihex', 'srec', 'auto', 'ti_txt', 'verilog_vmem', 'elf']:
         pass
     else:
         raise argparse.ArgumentTypeError(f"invalid input format '{fmt}'")
@@ -1709,6 +1829,8 @@ def _convert_output_format_type(value):
         args = (minimum_address, maximum_address)
     elif fmt == 'hexdump':
         pass
+    elif fmt == 'verilog_vmem':
+        pass
     else:
         raise argparse.ArgumentTypeError(f"invalid output format '{fmt}'")
 
@@ -1735,6 +1857,8 @@ def _do_convert_add_file(bf, input_format, infile, overwrite):
             bf.add_binary_file(infile, *args, overwrite=overwrite)
         elif fmt == 'ti_txt':
             bf.add_ti_txt_file(infile, *args, overwrite=overwrite)
+        elif fmt == 'verilog_vmem':
+            bf.add_verilog_vmem_file(infile, *args, overwrite=overwrite)
         elif fmt == 'elf':
             bf.add_elf_file(infile, *args, overwrite=overwrite)
     except AddDataError:
@@ -1755,6 +1879,8 @@ def _do_convert_as(bf, output_format):
         converted = bf.as_hexdump()
     elif fmt == 'ti_txt':
         converted = bf.as_ti_txt()
+    elif fmt == 'verilog_vmem':
+        converted = bf.as_verilog_vmem()
 
     return converted
 
@@ -1824,6 +1950,7 @@ def _do_as_hexdump(args):
         bf.add_file(binfile)
         print(bf.as_hexdump(), end='')
 
+
 def _do_as_ti_txt(args):
     for binfile in args.binfile:
         bf = BinFile()
@@ -1831,6 +1958,13 @@ def _do_as_ti_txt(args):
         print(bf.as_ti_txt(), end='')
 
 
+def _do_as_verilog_vmem(args):
+    for binfile in args.binfile:
+        bf = BinFile()
+        bf.add_file(binfile)
+        print(bf.as_verilog_vmem(), end='')
+
+
 def _do_fill(args):
     with open(args.infile, 'r') as fin:
         data = fin.read()
@@ -1901,14 +2035,14 @@ def _main():
         action='append',
         default=[],
         type=_convert_input_format_type,
-        help=('Input format auto, srec, ihex, ti_txt, elf, or binary[,<address>] '
-              '(default: auto). This argument may be repeated, selecting the '
-              'input format for each input file.'))
+        help=('Input format auto, srec, ihex, ti_txt, verilog_vmem, elf, or '
+              'binary[,<address>] (default: auto). This argument may be repeated, '
+              'selecting the input format for each input file.'))
     subparser.add_argument(
         '-o', '--output-format',
         default='hexdump',
         type=_convert_output_format_type,
-        help=('Output format srec, ihex, ti_txt, binary or hexdump '
+        help=('Output format srec, ihex, ti_txt, verilog_vmem, binary or hexdump '
               '(default: %(default)s).'))
     subparser.add_argument(
         '-s', '--word-size-bits',
@@ -1971,6 +2105,15 @@ def _main():
                            help='One or more binary format files.')
     subparser.set_defaults(func=_do_as_ti_txt)
 
+    # The 'as_verilog_vmem' subparser.
+    subparser = subparsers.add_parser(
+        'as_verilog_vmem',
+        description='Print given file(s) as Verilog VMEM.')
+    subparser.add_argument('binfile',
+                           nargs='+',
+                           help='One or more binary format files.')
+    subparser.set_defaults(func=_do_as_verilog_vmem)
+
     # The 'fill' subparser.
     subparser = subparsers.add_parser(
         'fill',
